C                                         Last modifications: 26.11.2009
c -----------------------------------------------------------
c Matrix x vector routine:  y = A*x
c -----------------------------------------------------------
      subroutine matmul(y, m,n,s,e,w, x, wrk, buf_out,buf_in,
     >                  me,north,south,east,west,lx,ly,bufsize)
      implicit none
      include 'mpif.h'
      integer i,j,lx,ly,me,north,south,east,west,bufsize
      integer ierr,stat(MPI_STATUS_SIZE)
      integer type_row,type_col,type_brow
      double precision m(lx,ly), 
     >                 n(lx,ly), s(lx,ly), 
     >                 e(lx,ly), w(lx,ly), 
     >                 x(lx,ly), y(lx,ly),
     >                 wrk(0:lx+1,0:ly+1),
     >                 buf_out(bufsize),buf_in(bufsize)

      COMMON/type_vect/ type_row,type_col,type_brow

c ----------------------------------------------------------------
       call vimbed(wrk,x,lx,ly)
       if (north.ne.-2) then
         do i=1,ly
            wrk(0,i)=wrk(1,i)
         enddo
       endif
       if (south.ne.-2) then
          do i=1,ly
	    wrk(lx+1,i)=wrk(lx,i)
          enddo
       endif
       if (east.ne.-2) then
          do i=1,lx
	    wrk(i,ly+1)=wrk(i,ly)
          enddo
       endif
       if (west.ne.-2) then
          do i=1,lx
            wrk(i,0)=wrk(i,1)
          enddo
       endif
       call MPI_BARRIER(MPI_COMM_WORLD,ierr)
c ----------------------------------------------------------------
c ------------- north-south exchange
       if (north.ne.-2) then
c       write(*,*) '1 >>>> ',me,' exchange with north'
           call MPI_SENDRECV(wrk(0,1),1,type_brow,north,1,
     >                       wrk(0,1),1,type_brow,north,2,
     >                       MPI_COMM_WORLD,stat,ierr)
       endif
       if (south.ne.-2) then
c       write(*,*) '1 >>>> ',me,' exchanges with south...'
         call MPI_SENDRECV(wrk(lx+1,1),1,type_brow,south,2,
     >                     wrk(lx+1,1),1,type_brow,south,1,
     >                     MPI_COMM_WORLD,stat,ierr)
       endif
c ----------------------------------------------------------------
c ------------- east-west exchange
       if (east.ne.-2) then
c       write(*,*) '3 >>>> ',me,' sends to east...'
           call MPI_SENDRECV(wrk(1,ly+1),1,type_col,east,3,
     >                       wrk(1,ly+1),1,type_col,east,4,
     >                       MPI_COMM_WORLD,stat,ierr)
       endif
       if (west.ne.-2) then
c       write(*,*) '4 >>>> ',me,' exchanges with west...'
          call MPI_SENDRECV(wrk(1,0),1,type_col,west,4,
     >                      wrk(1,0),1,type_col,west,3,
     >                      MPI_COMM_WORLD,stat,ierr)
       endif
       call MPI_BARRIER(MPI_COMM_WORLD,ierr)
c ----------------------------------------------------------------
c       write(*,*) 'Barrier passed','(',me,')'
c       write(*,*) 'end communicating','(',me,')'
c ----------------------------------------------------------------
       do i=1,lx
          do j=1,ly
             y(i,j) = m(i,j)*wrk(i,j)   + 
     >                e(i,j)*wrk(i+1,j) +
     >                w(i,j)*wrk(i-1,j) +
     >                n(i,j)*wrk(i,j+1) +
     >                s(i,j)*wrk(i,j-1)
          enddo
       enddo

      return
      end

c -----------------------------------------------------------
c Matrix generator routine:  
c 5-point stencil, central differences, homogeneous Dirichlet b.c.
c -----------------------------------------------------------
      subroutine matgen(m,n,s,e,w, 
     >           lx,ly,me,north,south,east,west)
      implicit none
      include 'mpif.h'
      integer i,j,lx,ly,north,south,east,west,me
      double precision m(lx,ly), 
     >                 n(lx,ly), s(lx,ly), 
     >                 e(lx,ly), w(lx,ly)

       do i=1,lx
         do j=1,ly
            m(i,j) =  4.0d0
            n(i,j) = -1.0d0
            s(i,j) = -1.0d0
            e(i,j) = -1.0d0
            w(i,j) = -1.0d0
         enddo
      enddo
c      write(*,*) 'matgen - full set','(',me,')'

c -------- find the global position of the subregion
      if (west .eq. -1) then
        do j=1,ly
            w(1,j) = 0.0d0
        enddo
      endif

      if (south .eq. -1) then
        do i=1,lx
            s(i,1) = 0.0d0
        enddo
      endif

      if (east .eq. -1) then
        do j=1,ly
            e(lx,j) = 0.0d0
        enddo
      endif

      if (north .eq. -1) then
        do i=1,lx
            n(i,ly) = 0.0d0
        enddo
      endif
c      call print(m,me,lx,ly,'Middle')
c      call print(n,me,lx,ly,'North')
c      call print(s,me,lx,ly,'South')
c      call print(e,me,lx,ly,'East')
c      call print(w,me,lx,ly,'West')

      return
      end
c -----------------------------------------------------------
c Right-hand-side routine:
c -----------------------------------------------------------
      subroutine rhsgen(rhs,u, m,n,s,e,w,
     >                  wrk, buf_out, buf_in,
     >                  me,north,south,east,west,
     >                  lx, ly, bufsize)

      implicit none
      integer i,j,lx,ly,bufsize
      integer north,south,east,west,me
      double precision rhs(lx,ly),u(lx,ly)
      real rand
      double precision m(lx,ly),
     >                 n(lx,ly), s(lx,ly),
     >                 e(lx,ly), w(lx,ly),
     >                 wrk(0:lx+1,0:ly+1),
     >                 buf_out(bufsize),buf_in(bufsize)
      do j=1,ly
         do i=1,lx
            u(i,j) = 1.0d0 !rand(0)
         enddo
      enddo
c      call print(u,me,lx,ly,'>>> the solution')
      call matmul(rhs, m,n,s,e,w, u, wrk,buf_out,buf_in,
     >            me,north,south,east,west,lx,ly,bufsize)

      return
      end

